@@ -1,25 +1,27 @@
 import os, math, urllib, re
-import toposort
 
 from dulwich.errors import HangupException
 from dulwich.index import commit_tree
 from dulwich.objects import Blob, Commit, Tag, Tree, parse_timezone
 from dulwich.pack import create_delta, apply_delta
 from dulwich.repo import Repo
+from dulwich import client
 
 from hgext import bookmarks
 from mercurial.i18n import _
 from mercurial.node import hex, bin, nullid
 from mercurial import context, util as hgutil
+from mercurial import error
 
+import util
 
 class GitHandler(object):
+    mapfile = 'git-mapfile'
+    tagsfile = 'git-tags'
 
     def __init__(self, dest_repo, ui):
         self.repo = dest_repo
         self.ui = ui
-        self.mapfile = 'git-mapfile'
-        self.tagsfile = 'git-tags'
 
         if ui.config('git', 'intree'):
             self.gitdir = self.repo.wjoin('.git')
@@ -148,7 +150,8 @@ class GitHandler(object):
 
         if remote_name and changed_refs:
             for ref, sha in changed_refs.iteritems():
-                self.ui.status("    "+ remote_name + "::" + ref + " => GIT:" + sha[0:8] + "\n")
+                self.ui.status("    %s::%s => GIT:%s\n" %
+                               (remote_name, ref, sha[0:8]))
 
             self.update_remote_branches(remote_name, changed_refs)
 
@@ -173,26 +176,23 @@ class GitHandler(object):
         nodes = [self.repo.lookup(n) for n in self.repo]
         export = [node for node in nodes if not hex(node) in self._map_hg]
         total = len(export)
-        if total:
-          magnitude = int(math.log(total, 10)) + 1
-        else:
-          magnitude = 1
         for i, rev in enumerate(export):
-            if i%100 == 0:
-                self.ui.status(_("at: %*d/%d\n") % (magnitude, i, total))
-
+            util.progress(self.ui, 'import', i, total=total)
             ctx = self.repo.changectx(rev)
             state = ctx.extra().get('hg-git', None)
             if state == 'octopus':
-                self.ui.debug("revision %d is a part of octopus explosion\n" % ctx.rev())
+                self.ui.debug("revision %d is a part "
+                              "of octopus explosion\n" % ctx.rev())
                 continue
             self.export_hg_commit(rev)
+        util.progress(self.ui, 'import', None, total=total)
+
 
     # convert this commit into git objects
     # go through the manifest, convert all blobs/trees we don't have
     # write the commit object (with metadata info)
     def export_hg_commit(self, rev):
-        self.ui.note(_("converting revision %s\n") % rev)
+        self.ui.note(_("converting revision %s\n") % hex(rev))
 
         oldenc = self.swap_out_encoding()
 
@@ -262,7 +262,7 @@ class GitHandler(object):
             author = author + ' <none@none>'
 
         if 'author' in ctx.extra():
-            author = apply_delta(author, ctx.extra()['author'])
+            author = "".join(apply_delta(author, ctx.extra()['author']))
 
         return author
 
@@ -290,7 +290,7 @@ class GitHandler(object):
 
         message = ctx.description() + "\n"
         if 'message' in extra:
-            message = apply_delta(message, extra['message'])
+            message = "".join(apply_delta(message, extra['message']))
 
         # HG EXTRA INFORMATION
         add_extras = False
@@ -354,53 +354,64 @@ class GitHandler(object):
         convert_list = {}
 
         # get a list of all the head shas
+        seenheads = set()
+        if refs is None:
+            refs = self.git.refs.as_dict()
         if refs:
-          for head, sha in refs.iteritems():
-              # refs contains all the refs in the server, not just the ones
-              # we are pulling
-              if sha in self.git.object_store:
-                  todo.append(sha)
-        else:
-            todo = self.git.refs.values()[:]
+            for sha in refs.itervalues():
+                # refs contains all the refs in the server, not just the ones
+                # we are pulling
+                if sha in self.git.object_store:
+                    obj = self.git.get_object(sha)
+                    while isinstance(obj, Tag):
+                        obj_type, sha = obj._get_object()
+                        obj = self.git.get_object(sha)
+                    if isinstance (obj, Commit) and sha not in seenheads:
+                        seenheads.add(sha)
+                        todo.append(sha)
+
+        # sort by commit date
+        def commitdate(sha):
+            obj = self.git.get_object(sha)
+            return obj.commit_time-obj.commit_timezone
+
+        todo.sort(key=commitdate, reverse=True)
 
         # traverse the heads getting a list of all the unique commits
+        commits = []
+        seen = set(todo)
         while todo:
-            sha = todo.pop()
-            assert isinstance(sha, str)
+            sha = todo[-1]
             if sha in done:
+                todo.pop()
                 continue
-            done.add(sha)
+            assert isinstance(sha, str)
             obj = self.git.get_object(sha)
-            if isinstance (obj, Commit):
+            assert isinstance(obj, Commit)
+            for p in obj.parents:
+                if p not in done:
+                    todo.append(p)
+                    break
+            else:
+                commits.append(sha)
                 convert_list[sha] = obj
-                todo.extend([p for p in obj.parents if p not in done])
-            if isinstance(obj, Tag):
-                (obj_type, obj_sha) = obj.get_object()
-                obj = self.git.get_object(obj_sha)
-                if isinstance (obj, Commit):
-                    convert_list[sha] = obj
-                    todo.extend([p for p in obj.parents if p not in done])
-
-        # sort the commits
-        commits = toposort.TopoSort(convert_list).items()
+                done.add(sha)
+                todo.pop()
 
         commits = [commit for commit in commits if not commit in self._map_git]
         # import each of the commits, oldest first
         total = len(commits)
-        if total:
-            magnitude = int(math.log(total, 10)) + 1
-        else:
-            magnitude = 1
         for i, csha in enumerate(commits):
-            if i%100 == 0:
-                self.ui.status(_("at: %*d/%d\n") % (magnitude, i, total))
+            util.progress(self.ui, 'import', i, total=total, unit='commits')
             commit = convert_list[csha]
             self.import_git_commit(commit)
+        util.progress(self.ui, 'import', None, total=total, unit='commits')
 
     def import_git_commit(self, commit):
         self.ui.debug(_("importing: %s\n") % commit.id)
 
-        (strip_message, hg_renames, hg_branch, extra) = self.extract_hg_metadata(commit.message)
+        (strip_message, hg_renames,
+         hg_branch, extra) = self.extract_hg_metadata(commit.message)
 
         # get a list of the changed, added, removed files
         files = self.get_files_changed(commit)
@@ -488,7 +499,8 @@ class GitHandler(object):
         if commit.author != commit.committer \
                or commit.author_time != commit.commit_time \
                or commit.author_timezone != commit.commit_timezone:
-            extra['committer'] = "%s %d %d" % (commit.committer, commit.commit_time, -commit.commit_timezone)
+            extra['committer'] = "%s %d %d" % (
+                commit.committer, commit.commit_time, -commit.commit_timezone)
 
         if commit.encoding:
             extra['encoding'] = commit.encoding
@@ -499,6 +511,16 @@ class GitHandler(object):
         if octopus:
             extra['hg-git'] ='octopus-done'
 
+        # TODO use 'n in self.repo' when we require hg 1.5
+        def repo_contains(n):
+            try:
+                return bool(self.repo.lookup(n))
+            except error.RepoLookupError:
+                return False
+
+        if not (repo_contains(p1) and repo_contains(p2)):
+            raise hgutil.Abort(_('you appear to have run strip - '
+                                 'please run hg git-cleanup'))
         ctx = context.memctx(self.repo, (p1, p2), text, list(files), getfilectx,
                              author, date, extra)
 
@@ -547,6 +569,30 @@ class GitHandler(object):
                 raise hgutil.Abort("revision %s cannot be pushed since"
                                    " it doesn't have a ref" % ctx)
 
+            # Check if the tags the server is advertising are annotated tags,
+            # by attempting to retrieve it from the our git repo, and building a
+            # list of these tags.
+            #
+            # This is possible, even though (currently) annotated tags are
+            # dereferenced and stored as lightweight ones, as the annotated tag
+            # is still stored in the git repo.
+            uptodate_annotated_tags = []
+            for r in tags:
+                ref = 'refs/tags/'+r
+                # Check tag.
+                if not ref in refs:
+                    continue
+                try:
+                    # We're not using Repo.tag(), as it's deprecated.
+                    tag = self.git.get_object(refs[ref])
+                    if not isinstance(tag, Tag):
+                        continue
+                except KeyError:
+                    continue
+
+                # If we've reached here, the tag's good.
+                uptodate_annotated_tags.append(ref)
+
             for r in heads + tags:
                 if r in heads:
                     ref = 'refs/heads/'+r
@@ -562,6 +608,9 @@ class GitHandler(object):
                     else:
                         raise hgutil.Abort("pushing %s overwrites %s"
                                            % (ref, ctx))
+                elif ref in uptodate_annotated_tags:
+                    # we already have the annotated tag.
+                    pass
                 else:
                     raise hgutil.Abort("%s changed on the server, please pull "
                                        "and merge before pushing" % ref)
@@ -578,11 +627,11 @@ class GitHandler(object):
                 for h in heads:
                     r = [ref for ref in refs if ref.endswith('/'+h)]
                     if not r:
-                        raise hgutil.Abort("ref %s not found on remote server")
+                        raise hgutil.Abort("ref %s not found on remote server" % h)
                     elif len(r) == 1:
                         want.append(refs[r[0]])
                     else:
-                        raise hgutil.Abort("ambiguous reference %s: %r"%(h, r))
+                        raise hgutil.Abort("ambiguous reference %s: %r" % (h, r))
             else:
                 want = [sha for ref, sha in refs.iteritems()
                         if not ref.endswith('^{}')]
@@ -590,7 +639,8 @@ class GitHandler(object):
         f, commit = self.git.object_store.add_pack()
         try:
             try:
-                return client.fetch_pack(path, determine_wants, graphwalker, f.write, self.ui.status)
+                return client.fetch_pack(path, determine_wants, graphwalker,
+                                         f.write, self.ui.status)
             except HangupException:
                 raise hgutil.Abort("the remote end hung up unexpectedly")
         finally:
@@ -644,7 +694,7 @@ class GitHandler(object):
                         sha = self.map_hg_get(refs[k])
                         self.tags[ref_name] = sha
                     elif isinstance (obj, Tag): # annotated
-                        (obj_type, obj_sha) = obj.get_object()
+                        (obj_type, obj_sha) = obj._get_object()
                         obj = self.git.get_object(obj_sha)
                         if isinstance (obj, Commit):
                             sha = self.map_hg_get(obj_sha)
@@ -688,24 +738,25 @@ class GitHandler(object):
                          ' bookmarks enabled?\n'))
 
     def update_remote_branches(self, remote_name, refs):
-        heads = dict([(ref[11:],refs[ref]) for ref in refs
-                      if ref.startswith('refs/heads/')])
-
-        for head, sha in heads.iteritems():
+        def _set_hg_tag(head, sha):
             # refs contains all the refs in the server, not just the ones
             # we are pulling
             if sha not in self.git.object_store:
-                continue
+                return
             hgsha = bin(self.map_hg_get(sha))
             tag = '%s/%s' % (remote_name, head)
             self.repo.tag(tag, hgsha, '', True, None, None)
 
-        for ref_name in refs:
+        for ref_name, sha in refs.iteritems():
             if ref_name.startswith('refs/heads'):
-                new_ref = 'refs/remotes/%s/%s' % (remote_name, ref_name[10:])
-                self.git.refs[new_ref] = refs[ref_name]
-            elif ref_name.startswith('refs/tags'):
-                self.git.refs[ref_name] = refs[ref_name]
+                head = ref_name[11:]
+                _set_hg_tag(head, sha)
+
+                new_ref = 'refs/remotes/%s/%s' % (remote_name, head)
+                self.git.refs[new_ref] = sha
+            elif (ref_name.startswith('refs/tags')
+                  and not ref_name.endswith('^{}')):
+                self.git.refs[ref_name] = sha
 
 
     ## UTILITY FUNCTIONS
@@ -765,6 +816,21 @@ class GitHandler(object):
         changes = self.git.object_store.tree_changes(btree, tree)
         files = {}
         for (oldfile, newfile), (oldmode, newmode), (oldsha, newsha) in changes:
+            # don't create new submodules
+            if newmode == 0160000:
+                if oldfile:
+                    # become a regular delete
+                    newfile, newmode = None, None
+                else:
+                    continue
+            # so old submodules shoudn't exist
+            if oldmode == 0160000:
+                if newfile:
+                    # become a regular add
+                    oldfile, oldmode = None, None
+                else:
+                    continue
+
             if newfile is None:
                 file = oldfile
                 delete = True
@@ -806,8 +872,9 @@ class GitHandler(object):
             return string.decode('ascii', 'replace').encode('utf-8')
 
     def get_transport_and_path(self, uri):
-        from dulwich.client import TCPGitClient, SSHGitClient, SubprocessGitClient
-        for handler, transport in (("git://", TCPGitClient), ("git@", SSHGitClient), ("git+ssh://", SSHGitClient)):
+        for handler, transport in (("git://", client.TCPGitClient),
+                                   ("git@", client.SSHGitClient),
+                                   ("git+ssh://", client.SSHGitClient)):
             if uri.startswith(handler):
                 # We need to split around : or /, whatever comes first
                 hostpath = uri[len(handler):]
@@ -829,4 +896,4 @@ class GitHandler(object):
                     transportpath = path
                 return transport(host, thin_packs=False), transportpath
         # if its not git or git+ssh, try a local url..
-        return SubprocessGitClient(thin_packs=False), uri
+        return client.SubprocessGitClient(thin_packs=False), uri
